/***************************************************************
 * Name:      drawing_canvas.cpp
 * Purpose:   For drawing music
 * Author:    John Conner Atkeson
 * Created:   2007-11-06
 * Copyright: John Conner Atkeson (http://jca3.freeshell.org)
 * License:   LGPL
 **************************************************************/

#include "drawing_canvas.h"

DrawingCanvasClass::DrawingCanvasClass(wxWindow *parent, int id) : wxPanel(parent, id, wxDefaultPosition, wxSize(-1, 30), wxSUNKEN_BORDER)
{
  m_parent = parent;
  this->SetSize(300, 300);

  // const long ID_PANEL_THIS = wxNewId();  Connect(ID_PANEL_THIS,wxEVT_PAINT,(wxObjectEventFunction)&DrawingCanvas::OnPaint,0,this);
  Connect(wxEVT_PAINT, wxPaintEventHandler(DrawingCanvasClass::OnPaint));
  // Connect(-1,wxEVT_PAINT,(wxObjectEventFunction)&DrawingCanvasClass::OnPaint,0,this);

  Connect(wxEVT_SIZE, wxSizeEventHandler(DrawingCanvasClass::OnSize));
  Connect(wxEVT_MOTION, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelMouseMove));

  Connect(wxEVT_LEFT_DOWN, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelLeftDown));
  Connect(wxEVT_LEFT_UP, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelLeftUp));

  Connect(wxEVT_RIGHT_DOWN, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelRightDown));
  Connect(wxEVT_RIGHT_UP, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelRightUp));

  // Connect(wxEVT_COMMAND_BUTTON_CLICKED, wxSizeEventHandler(DrawingCanvasClass::OnQuit));
  // Connect(ID_BUTTON2,wxEVT_COMMAND_BUTTON_CLICKED,(wxObjectEventFunction)&DrawingCanvasClass::OnQuit);
  // void OnShow(wxEvent& event); EVT_SHOW (MyFrame::OnShow);
  left_mouse_down = true;
  altstart = 0.0;
  loudness_start = 0.0;
#if have_player
  sgen = new Freewave_Generator_Class(440.0);

  sgen->freq = 440;
  player = new Player_Class();
#endif
  // Refresh();
  //Painter();
  //this->SetSize(300, 301);
  Voice_Track = new Voice_Track_Class(sgen, player);//new Freewave_Generator_Class(220.0)
  Voice_Track->fill_test();

  //player->slicer = sgen->spawn_slicer();
  //player->slicer = Voice_Track->spawn_slicer();
  player->slicer = NULL;
  selected_event = NULL;
  Box_Map = new Box_Mapper_Class();
  wxSize sz = GetClientSize();
  Box_Map->Set_Screen(0, sz.GetHeight(), sz.GetWidth(), 0);
  Box_Map->Set_Real(0,7, ((double)sz.GetWidth())/50.0, 12.0);
};
/* ************************************************************************************************************ */
DrawingCanvasClass::~DrawingCanvasClass(){
  delete Voice_Track;
  delete Box_Map;
#if have_player
  delete player;
  delete sgen;
#endif
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::Painter(wxDC& dc){
  // http://www.informit.com/articles/printerfriendly.aspx?p=405047&rl=1
  //wxPaintDC dc(this); //wxBufferedPaintDC
  wxSize sz = GetClientSize();
  // Box_Map->Set_Screen(0,0, sz.GetWidth(), sz.GetHeight());
  Box_Map->Set_Screen(0, sz.GetHeight(), sz.GetWidth(), 0);// inverted Y axis
  Box_Map->Set_Real(0,7, ((double)sz.GetWidth())/70.0, 12.0);
  {// blank background
    dc.SetBrush(*wxWHITE_BRUSH);
    dc.Clear();
  }
  // wxDouble x, y;
  // wg.MoveToPoint(x, y); wg.AddLineToPoint(x+1, y+1);
  wxPen pen(*wxRED, 1); // red pen of width 1
  dc.SetPen(pen);

  wxBrush pink_brush; pink_brush.SetColour(255, 0, 255);
  wxBrush red_brush; red_brush.SetColour(255, 0, 0);
  wxBrush grey_brush; grey_brush.SetColour(128, 128, 128);
  dc.SetBrush(pink_brush);

  // dc.DrawPoint(event.GetPosition());
  dc.DrawPoint(10, 10);
  dc.SetPen(wxNullPen);
  if (true){
    //wxPaintDC dc(DrawingPanel);
    dc.SetPen(*wxBLACK_PEN);
    // dc.SetBrush(*wxRED_BRUSH);
    // Get window dimensions
    wxSize sz = GetClientSize();

    // Our rectangle dimensions
    wxCoord w = 100, h = 50;
    // Center the rectangle on the window, but never
    // draw at a negative position.
    int x = wxMax(0, (sz.x-w)/2);
    int y = wxMax(0, (sz.y-h)/2);
    wxRect rectToDraw(x, y, w, h);
    // dc.DrawLine(x, y, w, h);
    // For efficiency, do not draw if not exposed
    if (IsExposed(rectToDraw)){
      dc.DrawRectangle(rectToDraw);
    }
  }
  //-------------------------------------------------------------------------------------
  // void draw_segment(int start_dex, int end_dex)
  if (true){
    int start_dex=0; int end_dex=Voice_Track->size()-1;
    Event_Marker_Class *event_prev = Voice_Track->get(start_dex);
    double x_prev, y_prev;
    double vol_prev = event_prev->loudness*10;
    Box_Map->Map_To_Screen(event_prev->time, event_prev->octave, &x_prev, &y_prev);

    for (int cnt=start_dex+1;cnt<=end_dex;cnt++){// inclusive
      Event_Marker_Class *event_next = Voice_Track->get(cnt);
      double x_next, y_next;
      double vol_next = event_next->loudness*10;
      Box_Map->Map_To_Screen(event_next->time, event_next->octave, &x_next, &y_next);

      dc.SetPen(*wxBLACK_PEN);
      if (false){
        dc.DrawLine((wxCoord)x_prev, (wxCoord)(y_prev-vol_prev), (wxCoord)x_next, (wxCoord)(y_next-vol_next));
        dc.DrawLine((wxCoord)x_prev, (wxCoord)(y_prev+vol_prev), (wxCoord)x_next, (wxCoord)(y_next+vol_next));
      }else{
        dc.SetBrush(grey_brush);
        wxPoint points[4];
        points[0].x=(int)x_prev; points[0].y=(int)(y_prev-vol_prev);
        points[1].x=(int)x_next; points[1].y=(int)(y_next-vol_next);
        points[2].x=(int)x_next; points[2].y=(int)(y_next+vol_next);
        points[3].x=(int)x_prev; points[3].y=(int)(y_prev+vol_prev);
        dc.DrawPolygon(4, points);

        // dc.SetPen(*wxTRANSPARENT_PEN); dc.SetPen(*wxBLACK_PEN);
        dc.SetBrush(red_brush);
        dc.DrawCircle((wxCoord)x_prev, (wxCoord)y_prev, 5);
      }
      event_prev = event_next;
      x_prev = x_next; y_prev = y_next; vol_prev = vol_next;
    }
    dc.SetBrush(red_brush); dc.DrawCircle((wxCoord)x_prev, (wxCoord)y_prev, 5);
  }
};/* DrawingCanvasClass::Painter */
#if 0
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelLeftDown(wxMouseEvent& event)
{
  printf("OnDrawingPanelLeftDown");
  left_mouse_down = true;
  bool altdown = event.AltDown();
  double xloc = event.GetX();
  xloc /= 50;
  // double yloc = event.GetY();
#if have_player
  player->quit(); player->sdlsoundinit();
  int event_num = Voice_Track->find_by_time(xloc);
  selected_event = Voice_Track->get(event_num);
  Voice_Track->play_segment(0, Voice_Track->size()-1);
#endif
  if (altdown){
    altstart = event.GetY();
    loudness_start = selected_event->loudness;
  }
};
#endif
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelRightDown(wxMouseEvent& event)
{
  printf("OnDrawingPanelRightDown");
  double xloc = event.GetX();
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelRightUp(wxMouseEvent& event)
{
  printf("OnDrawingPanelRightUp");
  double xloc = event.GetX();
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelLeftDown(wxMouseEvent& event)
{// from Waveform_Canvas_Class
  printf("OnDrawingPanelLeftDown");
  left_mouse_down = true;
  bool altdown = event.AltDown();
  double xloc = event.GetX();
  double yloc = event.GetY();
  double xray, yray;
  Box_Map->Map_To_Real(xloc,yloc, &xray,&yray);
#if have_player
  player->quit();

  // sgen->ampray[(int)xray] = yray;
  //    this should really be a part of mousedown;
  if (true){
    // detect the spline control point here, and move it.
    Event_Marker_Class *pnt = NULL;
    Event_Marker_Class *pnt_found = NULL;
    double pntxloc, pntyloc;
    int num_points = Voice_Track->size();
    for (int cnt=0;cnt<num_points;cnt++){
      pnt = Voice_Track->get(cnt);
      Box_Map->Map_To_Screen(pnt->time, pnt->octave, &pntxloc, &pntyloc);
      if (Math.Hypot(pntxloc-xloc,pntyloc-yloc)<5){
        pnt_found = pnt; break;
      }
    }
    selected_event = pnt_found;

    if (player->slicer!=NULL){
      player->quit();
      delete player->slicer; player->slicer = NULL;
    }
    if (selected_event!=NULL){
      player->slicer = selected_event->spawn_slicer();
      player->sdlsoundinit();
    }
    /*
    if (pnt_found!=NULL){
      selected_point = pnt_found;
    }
    */
  };
  //Voice_Track->play_segment(0, Voice_Track->size()-1);
#endif
  if (altdown){
    altstart = event.GetY();
    loudness_start = selected_event->loudness;
  }
};/* DrawingCanvasClass::OnDrawingPanelLeftDown */
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelLeftUp(wxMouseEvent& event)
{
  left_mouse_down = false;
#if have_player
  player->quit();
  if (player->slicer!=NULL){
    delete player->slicer; player->slicer = NULL;
  }
#endif
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelMouseMove(wxMouseEvent& event)
{
  //event.RightDown
  //bool buttonstate = event.GetButton();
  bool bdown_left = event.ButtonIsDown(wxMOUSE_BTN_LEFT);
  bool bdown_right = event.ButtonIsDown(wxMOUSE_BTN_RIGHT);
  bool altdown = event.AltDown();
  //bool dragging = event.Dragging();
  double xloc = event.GetX();
  double yloc = event.GetY();
  // bool buttonstate = event.Button(0);
  if (bdown_left){
    /*
    here we need to pick the exact event point of the voice and play it in a loop while we move it.
    alternatively we will be picking whole groups (any playable) and playing the whole group in a loop.
    */
    printf("left_mouse_down move!\n");
    if (selected_event!=NULL){
      double new_time,new_octave;
      Box_Map->Map_To_Real(xloc,yloc, &new_time, &new_octave);
      selected_event->time = new_time;
      // selected_event->time = xloc/50;
      if (altdown){
        // double new_loudness = loudness_start + (yloc-altstart)/10.0;
        // double new_loudness = (yloc-(selected_event->octave*5))/10.0;
        double new_loudness = new_octave-(selected_event->octave);

        if (new_loudness<0.0){
          new_loudness=0.0;
        }else if (new_loudness>1.0){
          new_loudness=1.0;
        }
        selected_event->loudness = new_loudness;
      }else{
        //selected_event->set_octave(new_octave);
        selected_event->set_octave(new_octave);
        //this->time=0.0; this->set_octave(7.0); this->loudness=1.0;;// defaults
        // selected_event->set_octave(8.0 + 0.01*(double)yloc);
      }
    }
    // handle graphics
    wxClientDC dc(this);
    wxSize clsz = this->GetClientSize();
    wxCoord w = 100, h = 50;
    int x = wxMax(0, (clsz.x-w)/2);
    int y = wxMax(0, (clsz.y-h)/2);
    // wxRect rectToDraw(x, y, w, h);
    // dc.DrawLine(x, y,  xloc, yloc);
    {
      // alternative way
      // wxBufferedDC bufdc(&dc, clsz, wxBUFFER_CLIENT_AREA);
      // Painter(bufdc); bufdc.DrawLine((wxCoord)x, (wxCoord)y,  (wxCoord)xloc, (wxCoord)yloc);
      int wdt = clsz.GetWidth(); int hgt = clsz.GetHeight();
      int bwdt = wdt, bhgt = hgt;

      wxMemoryDC memdc;
      // Draw to a memory device.
      wxBitmap newBitmap(bwdt, bhgt, -1);
      wxMemoryDC mem_dc;
      mem_dc.SelectObject(newBitmap);

      Painter(mem_dc);
      // mem_dc.DrawLine((wxCoord)x, (wxCoord)y,  (wxCoord)xloc, (wxCoord)yloc);

      int xsrc = 0, ysrc = 0;
      int xdest = 0, ydest = 0;
      dc.Blit(xdest, ydest, bwdt, bhgt, &mem_dc, xsrc, ysrc);
    }
  }else{
    printf("left_mouse_up move!\n");
  }
};/* DrawingCanvasClass::OnDrawingPanelMouseMove */
/* **********************************************************************************************************************
  Box_Mapper_Class methods
********************************************************************************************************************** */
Box_Mapper_Class::Box_Mapper_Class(){
  Set_Screen(0, 0, 1, 1);// default initializations.
  Set_Real(0, 0, 1, 1);
}
//-------------------------------------------------------------------
void Box_Mapper_Class::Set_Screen(double xleft, double ytop, double xright, double ybottom){
  scrleft = xleft; scrright = xright;
  scrtop = ytop; scrbottom = ybottom;
  scrwdt = xright-xleft; scrhgt = ybottom-ytop;
  set_ratios();
}
//-------------------------------------------------------------------
void Box_Mapper_Class::Set_Real(double xleft, double ytop, double xright, double ybottom){
  realleft = xleft; realright = xright;
  realtop = ytop; realbottom = ybottom;
  realwdt = xright-xleft; realhgt = ybottom-ytop;
  set_ratios();
}
//-------------------------------------------------------------------
void Box_Mapper_Class::set_ratios(){
  x_ratio_to_screen = scrwdt/realwdt; y_ratio_to_screen = scrhgt/realhgt;
}
//-------------------------------------------------------------------
double Box_Mapper_Class::Map_X_To_Screen(double xreal){
  return scrleft + ((xreal-realleft) * x_ratio_to_screen);
}
//-------------------------------------------------------------------
double Box_Mapper_Class::Map_Y_To_Screen(double yreal){
  return scrtop + ((yreal-realtop) * y_ratio_to_screen);
}
//-------------------------------------------------------------------
double Box_Mapper_Class::Map_X_To_Real(double xscr){
  return realleft + ((xscr-scrleft) / x_ratio_to_screen);
}
//-------------------------------------------------------------------
double Box_Mapper_Class::Map_Y_To_Real(double yscr){
  return realtop + ((yscr-scrtop) / y_ratio_to_screen);
}
//-------------------------------------------------------------------
void Box_Mapper_Class::Map_To_Screen(double xreal, double yreal, double *xscr, double *yscr){
  (*xscr) = Map_X_To_Screen(xreal); (*yscr) = Map_Y_To_Screen(yreal);
}
//-------------------------------------------------------------------
void Box_Mapper_Class::Map_To_Real(double xscr, double yscr, double *xreal, double *yreal){
  (*xreal) = Map_X_To_Real(xscr); (*yreal) = Map_Y_To_Real(yscr);
}
/* **********************************************************************************************************************
  Box_Mapper_Class methods
********************************************************************************************************************** */
