/***************************************************************
 * Name:      drawing_canvas.cpp
 * Purpose:   For drawing music
 * Author:    John Conner Atkeson
 * Created:   2007-11-06
 * Copyright: John Conner Atkeson (http://jca3.freeshell.org)
 * License:   LGPL
 **************************************************************/
#include "waveform_canvas.h"
Waveform_Canvas_Class::Waveform_Canvas_Class(wxWindow *parent, int id) : wxPanel(parent, id, wxDefaultPosition, wxSize(-1, 30), wxSUNKEN_BORDER)
{
  m_parent = parent;
  this->SetSize(500, 300);

  // const long ID_PANEL_THIS = wxNewId();  Connect(ID_PANEL_THIS,wxEVT_PAINT,(wxObjectEventFunction)&DrawingCanvas::OnPaint,0,this);
  Connect(wxEVT_PAINT, wxPaintEventHandler(Waveform_Canvas_Class::OnPaint));
  // Connect(-1,wxEVT_PAINT,(wxObjectEventFunction)&Waveform_Canvas_Class::OnPaint,0,this);

  Connect(wxEVT_SIZE, wxSizeEventHandler(Waveform_Canvas_Class::OnSize));
  Connect(wxEVT_MOTION, wxMouseEventHandler(Waveform_Canvas_Class::OnDrawingPanelMouseMove));

  Connect(wxEVT_LEFT_DOWN, wxMouseEventHandler(Waveform_Canvas_Class::OnDrawingPanelLeftDown));
  Connect(wxEVT_LEFT_UP, wxMouseEventHandler(Waveform_Canvas_Class::OnDrawingPanelLeftUp));

  // Connect(wxEVT_COMMAND_BUTTON_CLICKED, wxSizeEventHandler(Waveform_Canvas_Class::OnQuit));
  // Connect(ID_BUTTON2,wxEVT_COMMAND_BUTTON_CLICKED,(wxObjectEventFunction)&Waveform_Canvas_Class::OnQuit);
  // void OnShow(wxEvent& event); EVT_SHOW (MyFrame::OnShow);
  left_mouse_down = true;
  altstart = 0.0;
  //loudness_start = 0.0;
#if have_player
  player = new Player_Class();
  double default_freq;
  //default_freq = 220.0;
  //default_freq = 440.0;
  //default_freq = 262.0;
  default_freq = 110.0;
  // sgen = new Signal_Generator_Class();
  // sgen = new Freewave_Generator_Class(default_freq);
  sgen = new Splinewave_Generator_Class(default_freq);
  // this->Link_Signal_Generator(sgen);
  // player->sgen = sgen;
#endif
  // Refresh();
  //Painter();
  //this->SetSize(300, 301);

  Voice_Track = new Voice_Track_Class(sgen, player);
  //new Freewave_Generator_Class(220.0)
#if 0
  Voice_Track->fill_test();
  player->slicer = sgen->spawn_slicer();
#else
  cpoint = new Event_Marker_Class();
  cpoint->time = 0.5;
  cpoint->set_frequency(default_freq); // cpoint->set_octave(Math.Log(default_freq)/Math.Log(2.0));
  cpoint->loudness = 1.0;
  Voice_Track->add(cpoint);
  player->slicer = cpoint->spawn_slicer();
#endif

  // player->slicer = Voice_Track->spawn_slicer();

  // selected_event = NULL;
  selected_point = NULL;
  Box_Map = new Box_Mapper_Class();
  Spline_Box_Map = new Box_Mapper_Class();
  wxSize sz = GetClientSize();
  Box_Map->Set_Screen(0, sz.GetHeight(), sz.GetWidth(), 0);
  Box_Map->Set_Real(0,7, ((double)sz.GetWidth())/50.0, 12.0);
};
/* ************************************************************************************************************ */
Waveform_Canvas_Class::~Waveform_Canvas_Class(){
  delete Voice_Track;
  delete Box_Map;
  delete Spline_Box_Map;
#if have_player
  delete player;
  delete sgen;
#endif
};
interface I {
  //public void doSomething();
};
#if 0
void xyz(){
  Create(parent, id, _("wxWidgets app"), wxDefaultPosition, wxDefaultSize, wxDEFAULT_DIALOG_STYLE|wxRESIZE_BORDER, _T("wxDialog"));
  // Slider1 = new wxSlider(this, ID_SLIDER1, 0, 0, 100, wxDefaultPosition, wxDefaultSize, 0, wxDefaultValidator, _T("ID_SLIDER1"));
  Slider1 = new wxSlider(this, ID_SLIDER1, 0, 0, 100, wxDefaultPosition, wxDefaultSize, wxSL_VERTICAL|wxSL_INVERSE, wxDefaultValidator, _T("ID_SLIDER1"));
  BoxSizer1->Add(Slider1, 0, wxALL|wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL, 0);

  Connect(ID_SLIDER1,wxEVT_SCROLL_TOP|wxEVT_SCROLL_BOTTOM|wxEVT_SCROLL_LINEUP|wxEVT_SCROLL_LINEDOWN|wxEVT_SCROLL_PAGEUP|wxEVT_SCROLL_PAGEDOWN|wxEVT_SCROLL_THUMBTRACK|wxEVT_SCROLL_THUMBRELEASE|wxEVT_SCROLL_CHANGED,(wxObjectEventFunction)&exp_wxDialog::OnSlider1CmdScroll);
  Connect(ID_SLIDER1,wxEVT_SCROLL_CHANGED,(wxObjectEventFunction)&exp_wxDialog::OnSlider1CmdScrollChanged);
}
void exp_wxDialog::OnSlider1CmdScrollChanged(wxScrollEvent& event)
{
  wxString msgx;

  char msg[300];
  double pos = event.GetPosition();
  // sprintf(msg,"GetPosition:%lf:",pos);
  msgx = wxString::Format(wxT("GetPosition:%lf:"),pos);
  wxMessageBox(msgx);
}

void exp_wxDialog::OnSlider1CmdScroll(wxScrollEvent& event)
{
  wxMessageBox(wxT("Scroll!"));
}
#endif
/* ************************************************************************************************************ */
void Waveform_Canvas_Class::Painter(wxDC& dc){
  // http://www.informit.com/articles/printerfriendly.aspx?p=405047&rl=1
  //wxPaintDC dc(this); //wxBufferedPaintDC
  wxSize sz = GetClientSize();
  // Box_Map->Set_Screen(0,0, sz.GetWidth(), sz.GetHeight());
  Box_Map->Set_Screen(0, sz.GetHeight(), sz.GetWidth(), 0);// inverted Y axis
  Box_Map->Set_Real(-2,-1.0, sgen->frez, 1.0);

  Spline_Box_Map->Set_Screen(0, sz.GetHeight(), sz.GetWidth(), 0);// inverted Y axis
  Spline_Box_Map->Set_Real(0.0,-1.0, 1.0, 1.0);
  {// blank background
    dc.SetBrush(*wxWHITE_BRUSH);
    dc.Clear();
  }
  // wxDouble x, y;
  // wg.MoveToPoint(x, y); wg.AddLineToPoint(x+1, y+1);
  wxPen pen(*wxRED, 1); // red pen of width 1
  dc.SetPen(pen);

  wxBrush pink_brush; pink_brush.SetColour(255, 0, 255);
  wxBrush red_brush; red_brush.SetColour(255, 0, 0);
  wxBrush grey_brush; grey_brush.SetColour(128, 128, 128);
  dc.SetBrush(pink_brush);

  // dc.DrawPoint(event.GetPosition());
  dc.DrawPoint(10, 10);
  dc.SetPen(wxNullPen);
  if (true){
    //wxPaintDC dc(DrawingPanel);
    dc.SetPen(*wxBLACK_PEN);
    // dc.SetBrush(*wxRED_BRUSH);
    // Get window dimensions
    wxSize sz = GetClientSize();

    // Our rectangle dimensions
    wxCoord w = 100, h = 50;
    // Center the rectangle on the window, but never
    // draw at a negative position.
    int x = wxMax(0, (sz.x-w)/2);
    int y = wxMax(0, (sz.y-h)/2);
    wxRect rectToDraw(x, y, w, h);
    // dc.DrawLine(x, y, w, h);
    // For efficiency, do not draw if not exposed
    if (IsExposed(rectToDraw)){
      dc.DrawRectangle(rectToDraw);
    }
  }
  //-------------------------------------------------------------------------------------
  if (true){
    int rez = sgen->rez;
    double x_prev, y_prev;
    double x_next, y_next;
    x_prev = -1;
    y_prev = sgen->ampray[rez-1];
    Box_Map->Map_To_Screen(x_prev, y_prev, &x_prev, &y_prev);
    for (int cnt=0;cnt<rez;cnt++){
      x_next = ((double)cnt);
      y_next = sgen->ampray[cnt];
      Box_Map->Map_To_Screen(x_next, y_next, &x_next, &y_next);
      dc.DrawLine((wxCoord)x_prev, (wxCoord)y_prev, (wxCoord)x_next, (wxCoord)y_next);
      x_prev = x_next; y_prev = y_next;
    }
  }
  if (true){
    dc.SetBrush(red_brush);
    int num_points = sgen->spline->size();
    double x_prev, y_prev;
    double x_next, y_next;

    Splinewave_Generator_Class::Point_Class *pnt = sgen->spline->get(num_points-1);
    x_prev = pnt->xloc-1.0; y_prev = pnt->yloc;
    Spline_Box_Map->Map_To_Screen(x_prev, y_prev, &x_prev, &y_prev);
    for (int cnt=0;cnt<num_points;cnt++){
      int pdex = cnt%num_points;
      pnt = sgen->spline->get(pdex); x_next = pnt->xloc; y_next = pnt->yloc;
      Spline_Box_Map->Map_To_Screen(x_next, y_next, &x_next, &y_next);
      dc.SetPen(*wxRED_PEN);
      dc.DrawLine((wxCoord)x_prev, (wxCoord)y_prev, (wxCoord)x_next, (wxCoord)y_next);
      dc.SetPen(*wxBLACK_PEN);
      dc.DrawCircle((wxCoord)x_next, (wxCoord)y_next, 5);
      x_prev = x_next; y_prev = y_next;
    }
    pnt = sgen->spline->get(0); x_next = pnt->xloc+1.0; y_next = pnt->yloc;
    Spline_Box_Map->Map_To_Screen(x_next, y_next, &x_next, &y_next);
    dc.SetPen(*wxRED_PEN);
    dc.DrawLine((wxCoord)x_prev, (wxCoord)y_prev, (wxCoord)x_next, (wxCoord)y_next);
  }
};/* Waveform_Canvas_Class::Painter */
/* ************************************************************************************************************ */
void Waveform_Canvas_Class::OnDrawingPanelLeftDown(wxMouseEvent& event)
{
  printf("OnDrawingPanelLeftDown");
  left_mouse_down = true;
  bool altdown = event.AltDown();
  double xloc = event.GetX();
  double yloc = event.GetY();
  double xray, yray;
  Box_Map->Map_To_Real(xloc,yloc, &xray,&yray);
#if have_player
  player->quit(); player->sdlsoundinit();
  // sgen->ampray[(int)xray] = yray;
  //    this should really be a part of mousedown;
  if (true){
    // detect the spline control point here, and move it.
    Splinewave_Generator_Class::Point_Class *pnt = NULL;
    Splinewave_Generator_Class::Point_Class *pnt_found = NULL;
    double pntxloc, pntyloc;
    int num_points = sgen->spline->size();
    for (int cnt=0;cnt<num_points;cnt++){
      pnt = sgen->spline->get(cnt);
      Spline_Box_Map->Map_To_Screen(pnt->xloc, pnt->yloc, &pntxloc, &pntyloc);
      if (Math.Hypot(pntxloc-xloc,pntyloc-yloc)<5){
        pnt_found = pnt; break;
      }
    }
    selected_point = pnt_found;
    /*
    if (pnt_found!=NULL){
      selected_point = pnt_found;
    }
    */
  }
  //Voice_Track->play_segment(0, Voice_Track->size()-1);
#endif
  if (altdown){
    altstart = event.GetY();
    //loudness_start = selected_event->loudness;
  }
}
/* ************************************************************************************************************ */
void Waveform_Canvas_Class::OnDrawingPanelMouseMove(wxMouseEvent& event)
{
  bool bdown = event.ButtonIsDown(wxMOUSE_BTN_LEFT);
  //bool altdown = event.AltDown();
  //bool dragging = event.Dragging();
  double xloc = event.GetX();
  double yloc = event.GetY();
  // bool buttonstate = event.Button(0);
  if (bdown){
    printf("left_mouse_down move!\n");
    double xray, yray;
    Box_Map->Map_To_Real(xloc,yloc, &xray, &yray);
    if (xray<0){
      xray=0;
    } else if (xray>=sgen->rez){
      xray=sgen->rez-1;
    }
//    this should really be a part of mousedown;
    if (true){
      // detect the spline control point here, and move it.
      if (selected_point!=NULL){
        Spline_Box_Map->Map_To_Real(xloc, yloc, &(selected_point->xloc), &(selected_point->yloc));
        sgen->apply_spline();
      }
    }
  }
  if (bdown){
    wxClientDC dc(this);
    wxSize clsz = this->GetClientSize();
    // clsz.DecBy(10, 10);

    wxCoord w = 100, h = 50;
    int x = wxMax(0, (clsz.x-w)/2);
    int y = wxMax(0, (clsz.y-h)/2);
    // wxRect rectToDraw(x, y, w, h);
    // dc.DrawLine(x, y,  xloc, yloc);
    {
      //wxBufferedDC bufdc(&dc, clsz, wxBUFFER_CLIENT_AREA);
      //Painter(bufdc); bufdc.DrawLine((wxCoord)x, (wxCoord)y,  (wxCoord)xloc, (wxCoord)yloc);

      int wdt = clsz.GetWidth(); int hgt = clsz.GetHeight();
      int bwdt = wdt, bhgt = hgt;

      wxMemoryDC memdc;
      // Draw to a memory device.
      wxBitmap newBitmap(bwdt, bhgt, -1);
      wxMemoryDC mem_dc;
      mem_dc.SelectObject(newBitmap);

      Painter(mem_dc);
      // mem_dc.DrawLine((wxCoord)x, (wxCoord)y,  (wxCoord)xloc, (wxCoord)yloc);

      int xsrc = 0, ysrc = 0;
      int xdest = 0, ydest = 0;
      dc.Blit(xdest, ydest, bwdt, bhgt, &mem_dc, xsrc, ysrc);
    }
  }else{
    printf("left_mouse_up move!\n");
  }
};/* Waveform_Canvas_Class::OnDrawingPanelMouseMove */
/* **********************************************************************************************************************
  Begin Waveform_Editor_Class methods
********************************************************************************************************************** */
Waveform_Editor_Class::Waveform_Editor_Class(wxWindow *parent, int id) : wxPanel(parent, id, wxDefaultPosition, wxSize(-1, 30), wxSUNKEN_BORDER){
  BoxSizer = new wxBoxSizer(wxHORIZONTAL);

  const long ID_PANEL_WC = wxNewId();
  Waveform_Canvas = new Waveform_Canvas_Class(this, ID_PANEL_WC);
  BoxSizer->Add(Waveform_Canvas, 0, wxALL|wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL, 0);

  const long ID_OCTAVE_SLIDER = wxNewId();
  Octave_Slider = new wxSlider(this, ID_OCTAVE_SLIDER, 0, 0, 100, wxDefaultPosition, wxSize(-1,300), wxSL_VERTICAL|wxSL_INVERSE, wxDefaultValidator, wxT("ID_OCTAVE_SLIDER"));
  BoxSizer->Add(Octave_Slider, 0, wxALL|wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL, 0);

  //Octave_Slider->SetSize(-1, 900);//wxSize(-1,900) wxDefaultSize
  int value=6000;
  Octave_Slider->SetMin(4000); Octave_Slider->SetMax(13000);
  Octave_Slider->SetValue(value);
  Octave_Slider->Refresh();

  const long ID_SAVE_BUTTON = wxNewId();
  Save_Button = new wxButton(this, ID_SAVE_BUTTON, _("Save"), wxPoint(17,70), wxDefaultSize, 0, wxDefaultValidator, wxT("ID_SAVE_BUTTON"));
  BoxSizer->Add(Save_Button, 0, wxALL|wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL, 0);

  SetSizer(BoxSizer);
  BoxSizer->Fit(this);

  Connect(ID_OCTAVE_SLIDER,wxEVT_SCROLL_CHANGED,(wxObjectEventFunction)&Waveform_Editor_Class::OnOctave_SliderCmdScrollChanged);
  Connect(ID_OCTAVE_SLIDER,wxEVT_SCROLL_CHANGED,(wxObjectEventFunction)&Waveform_Editor_Class::OnOctave_SliderCmdScrollChanged);
  Connect(ID_OCTAVE_SLIDER,wxEVT_SCROLL_THUMBTRACK,(wxObjectEventFunction)&Waveform_Editor_Class::OnOctave_SliderCmdScrollThumbTrack);
  Connect(ID_OCTAVE_SLIDER,wxEVT_SCROLL_THUMBRELEASE,(wxObjectEventFunction)&Waveform_Editor_Class::OnOctave_SliderCmdScrollThumbRelease);

  Connect(ID_SAVE_BUTTON,wxEVT_COMMAND_BUTTON_CLICKED,(wxObjectEventFunction)&Waveform_Editor_Class::OnButton1Click);
}
/* ************************************************************************************************************ */
void Waveform_Editor_Class::OnOctave_SliderCmdScrollChanged(wxScrollEvent& event)
{
  double pos;
  pos = event.GetInt();//pos = event.GetPosition();
  pos /= 1000.0;
  Waveform_Canvas->sgen->set_octave(pos);
};
/* ************************************************************************************************************ */
void Waveform_Editor_Class::OnOctave_SliderCmdScroll(wxScrollEvent& event)
{
  bool noop = true;
};
/* ************************************************************************************************************ */
void Waveform_Editor_Class::OnOctave_SliderCmdScrollThumbTrack(wxScrollEvent& event)
{
  double pos;
  pos = event.GetInt();
  //pos = event.GetPosition();
  pos /= 1000;
  Waveform_Canvas->cpoint->set_octave(pos);
  // now we want the sound to play every time we touch this control.
  if (!Waveform_Canvas->player->Initialised){
    Waveform_Canvas->player->sdlsoundinit();
  }
};
/* ************************************************************************************************************ */
void Waveform_Editor_Class::OnOctave_SliderCmdScrollLeftDown(wxScrollEvent& event)
{
#if 0
  Waveform_Canvas->player->quit();
  Waveform_Canvas->player->sdlsoundinit();
#endif
};
/* ************************************************************************************************************ */
void Waveform_Editor_Class::OnOctave_SliderCmdScrollThumbRelease(wxScrollEvent& event)
{
  Waveform_Canvas->player->quit();
};
/* ************************************************************************************************************ */
void Waveform_Editor_Class::OnButton1Click(wxCommandEvent& event)
{
  FILE *outfile = fopen("outfile.txt","w");
  int num_points = Waveform_Canvas->sgen->spline->size();
  for (int pcnt=0;pcnt<num_points;pcnt++){
    Splinewave_Generator_Class::Point_Class *pnt = Waveform_Canvas->sgen->spline->get(pcnt);
    fprintf(outfile,"{%lf, %lf} ",pnt->xloc,pnt->yloc);
  }
  fprintf(outfile,"\n");
  fclose(outfile);
};
/* ************************************************************************************************************ */

